"""Return current version, this should be removed."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

__version__ = "1.2.1"


def get_version():
    """Return version."""
    return __version__
