"""Initial file for mmdetection hooks."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .det_class_probability_map_hook import DetClassProbabilityMapHook

__all__ = ["DetClassProbabilityMapHook"]
