"""Builder module for otx.core.ov.graph.parsers."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from otx.core.ov.registry import Registry

PARSERS = Registry("ov graph parsers")
