"""OTX Adapters for action recognition backbones - mmaction2."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .movinet_recognizer import MoViNetRecognizer

__all__ = ["MoViNetRecognizer"]
