"""Module for otx.core.ov.graph."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

# TODO: Need to remove comment with ignore mypy and fix mypy issues
from .graph import Graph  # type: ignore[attr-defined]

__all__ = ["Graph"]
