"""Neck list of mmdetection adapters."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .mmov_fpn import MMOVFPN
from .mmov_ssd_neck import MMOVSSDNeck
from .mmov_yolov3_neck import MMOVYOLOV3Neck

__all__ = [
    "MMOVFPN",
    "MMOVSSDNeck",
    "MMOVYOLOV3Neck",
]
