"""ROI Extractor list for mmdetection adapters."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .single_level_roi_extractor import SingleRoIExtractor

__all__ = ["SingleRoIExtractor"]
